#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "norm.h"

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <errno.h>

#include <norm_debug.h>
#include <norm_ip.h>
#include <norm_icmp.h>
#include <norm_misc.h>
#include <norm_options.h>
#include <norm_stats.h>

#ifdef LINUX
#define __FAVOR_BSD
#endif

int
norm_icmp_checksum_valid(struct icmp *icmphdr, int icmp_len)
{
  unsigned short  origsum;
  unsigned short  newsum;

  D_ENTER;

  origsum = icmphdr->icmp_cksum;
  icmphdr->icmp_cksum = 0;

  newsum  = norm_ip_in_cksum((u_short *) icmphdr, icmp_len, 0);
  D(NOISE_LOGLEVEL, ("ICMP checksum: %u -- %u\n", newsum, origsum));

  if (newsum == origsum)
    {
      icmphdr->icmp_cksum = origsum;
      D_RETURN_(1);
    }

  icmphdr->icmp_cksum = origsum;
  D_RETURN_(0);
}


int   
norm_icmp_handler(NormPacket *p, Interface *nif)
{
  struct ip      *iphdr;
  struct icmp    *icmphdr;
  int             ip_length;

  D_ENTER;

  iphdr     = (struct ip*)p->network_data;
  icmphdr   = (struct icmp*)p->transport_data;
  ip_length = norm_packet_get_length_ip(p);

  switch (icmphdr->icmp_type)
    {
    case ICMP_ECHOREPLY:       /* echo reply */
      print_icmp(INFO_LOGLEVEL, ("Echo reply.\n"));
      break;
    case ICMP_UNREACH:           /* dest unreachable */
      print_icmp(INFO_LOGLEVEL, ("Destination unreachable.\n"));
      break;
    case ICMP_SOURCEQUENCH:  /* packet lost, slow down */
      if (options.icmp_drop_sourcequench)
	{
	  print_icmp(NORM_LOGLEVEL, ("Source quench message -- dropping.\n"));
	  D_RETURN_(-1);
	}
      break;
    case ICMP_REDIRECT:      /* shorter route, codes: */
      print_icmp(INFO_LOGLEVEL, ("Redirect.\n"));
      break;
    case ICMP_ECHO:          /* echo service */
      print_icmp(INFO_LOGLEVEL, ("Echo request.\n"));
      break;
    case ICMP_ROUTERADVERT:  /* router advertisement */
      print_icmp(INFO_LOGLEVEL, ("Router advert..\n"));
      break;
    case ICMP_ROUTERSOLICIT: /* router solicitation */
      print_icmp(INFO_LOGLEVEL, ("Router solict.\n"));
      break;
    case ICMP_TIMXCEED:      /* time exceeded, code: */
      print_icmp(INFO_LOGLEVEL, ("Time exceeded.\n"));
      break;
    case ICMP_PARAMPROB:     /* ip header bad */
      print_icmp(INFO_LOGLEVEL, ("Bad IP header.\n"));
      break;
    case ICMP_TSTAMP:        /* timestamp request */
      print_icmp(INFO_LOGLEVEL, ("Timestamp req.\n"));
      break;
    case ICMP_TSTAMPREPLY:   /* timestamp reply */
      print_icmp(INFO_LOGLEVEL, ("Timestamp rep.\n"));
      break;
    case ICMP_IREQ:          /* information request */
      print_icmp(INFO_LOGLEVEL, ("Info req.\n"));
      break;
    case ICMP_IREQREPLY:     /* information reply */
      print_icmp(INFO_LOGLEVEL, ("Info rep.\n"));
      break;
    case ICMP_MASKREQ:       /* address mask request */
      print_icmp(INFO_LOGLEVEL, ("Address mask req.\n"));
      break;
    case ICMP_MASKREPLY:     /* address mask reply */
      print_icmp(INFO_LOGLEVEL, ("Address mask rep.\n"));
      break;
    default:
    }

  if (options.icmp_verify_checksum)
    {
      if (!norm_icmp_checksum_valid(icmphdr, ip_length - iphdr->ip_hl*4))
	{
	  norm_stats_increment_icmp_norms();
	  print_icmp(NORM_LOGLEVEL, ("Packet has invalid checksum -- dropping.\n"));
	  D_RETURN_(-1);
	}
    }

  if (options.icmp_drop_nonzero_code)
    {
      if (icmphdr->icmp_code != 0)
	{
	  norm_stats_increment_icmp_norms();
	  print_icmp(NORM_LOGLEVEL, ("Packet has non-zero code field -- dropping.\n"));
	  D_RETURN_(-1);
	}
    }

  if (options.icmp_drop_broadcast)
    {
      /* If the packet is coming from the external side and is a ping request ... */
      if ((!norm_misc_is_trusted_address(iphdr->ip_src)) &&
	  (icmphdr->icmp_type == ICMP_ECHO))
	{
	  /* FIX THIS FINALLY! */

	  /*
	  struct ifreq ifr;
	  strcpy(ifr.ifr_name, nifs[INTERNAL_IF].device);

	  if (ioctl(sock, SIOCGIFBRDADDR, &ifr) != -1)
	    {	  
	      if (htonl(iphdr->ip_dst.s_addr) == htonl(((struct sockaddr_in *)&ifr.ifr_broadaddr)->sin_addr.s_addr))
		{
		  print_icmp(("Ping request to net-directed broadcast -- dropping.\n"));
		  D_RETURN_(-1);
		}
	    }
	  */
	}
    }

  if (options.icmp_drop_multicast && icmphdr->icmp_type == ICMP_ECHO)
    {
      /* Check if it's a multicast address. Compare uppermost nibble of IP addr. to 16. */
      if (((htonl(iphdr->ip_dst.s_addr) & 0xf0000000) >> 28) == 16)
	{
	  norm_stats_increment_icmp_norms();
	  print_icmp(NORM_LOGLEVEL, ("Ping request to multicast address -- dropping.\n"));
	  D_RETURN_(-1);
	}
    }
  
  D_RETURN_(0);
  nif = NULL;
}
