#ifndef __norm_tcb_h
#define __norm_tcb_h

#ifdef LINUX
#define __FAVOR_BSD
#endif
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include <norm_list.h>

#define NORM_TCB_HASH_SIZE 257
#define NORM_TCB_SLOT_SIZE 100

/* TCP Control Block - holds all our state about a particular
 * connection. We have a notion of the internal end and the
 * external end of the connection as viewed from the point
 * of view of the normalizer.
*/

typedef struct norm_tcp_packet NormTCPPacket;
typedef struct norm_tcb        NormTCB;

struct norm_tcp_packet
{
  u_char      *data;
  u_int        len;
  u_int32_t    seq_no;
};


struct norm_tcb
{
  u_short   int_port, ext_port;
  u_int32_t int_addr, ext_addr;
  
  u_int32_t first_int_seq_no;
  u_int32_t first_ext_seq_no;

  u_int32_t int_ack_no;            /* Highest ack by internal host */
  u_int32_t ext_ack_no;            /* Highest ack by external host */
  u_int32_t int_seq_no;            /* Highest seqno sent by internal host
				      and forwarded by the normalizer           */
  u_int32_t ext_seq_no;            /* Highest seqno sent by internal host
				      and forwarded by the normalizer           */

  u_int16_t int_win;               /* Window size of the internal host, as last
				      advertised, 0 otherwise.
				   */

  /* first packet sent by the internal host that we've stored because 
     we've not yet got an ACK for it*/
  /* this is a linked list of stored packets in decending seq_no order*/
  NormList *int_packets;
  int       num_int_packets;

  /* first packet sent by the external host that we've stored because 
     we've not yet got an ACK for it*/
  /* this is a linked list of stored packets in decending seq_no order*/
  NormList *ext_packets; 
  int       num_ext_packets;

  u_int8_t  ext_last_byte;         /* last byte sent by external host and acked
				      by internal host */
  int       ext_last_byte_valid : 1;

  int       int_syn_seen : 1;
  int       ext_syn_seen : 1;
  int       int_fin_seen : 1;
  int       ext_fin_seen : 1;
  int       int_rst_seen : 1;
  int       ext_rst_seen : 1;
  
  /* Have we seen packets in each direction for this flow? */
  /* We can use this to discard spurious state in DoS attacks. */
  /* Names mean interface side. */
  int       seen_int     : 1;
  int       seen_ext     : 1;
                                 
};


/* Initalization of TCP state management */
void      norm_tcb_init(void);

/*Instantiate a TCB for a new flow*/
NormTCB  *norm_tcb_new(struct ip *ih, struct tcphdr *th,
		       Interface *nif);

void      norm_tcb_free(NormTCB *tcb);

NormTCB  *norm_tcb_lookup(struct ip *ih, struct tcphdr *th, Interface *nif);

int       norm_tcb_update(NormTCB *tcb, NormPacket *p, Interface *nif);

/* Adjust a segment to the current window -- returns value < 0 if
 * there is no overlap, or no window available in the TCB.
 */
int       norm_tcb_check_window(NormTCB *tcb, NormPacket *p, Interface *nif);

void      norm_tcb_store(NormTCB *tcb);
void      norm_tcb_flush(NormTCB *tcb);

#endif
